# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""TemplateFlow is the Zone of Templates."""
__packagename__ = "templateflow"
__copyright__ = "2020, The TemplateFlow developers"
try:
    from ._version import __version__
except ModuleNotFoundError:
    from pkg_resources import get_distribution, DistributionNotFound
    try:
        __version__ = get_distribution(__packagename__).version
    except DistributionNotFound:
        __version__ = "unknown"
    del get_distribution
    del DistributionNotFound

import os
from . import api
from .conf import update, TF_USE_DATALAD


if not TF_USE_DATALAD and os.getenv("TEMPLATEFLOW_AUTOUPDATE", "1") not in (
    "false",
    "off",
    "0",
    "no",
    "n",
):
    # trigger skeleton autoupdate
    update(local=True, overwrite=False, silent=True)

__all__ = [
    '__copyright__',
    '__packagename__',
    '__version__',
    'api',
    'update',
]



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Check S3-type repo tooling."""
# import pytest
from pathlib import Path
import requests
from .. import _s3 as s3


def test_get_skel_file(monkeypatch):
    """Exercise the skeleton file generation."""
    local_md5 = s3.TF_SKEL_MD5
    monkeypatch.setattr(s3, "TF_SKEL_MD5", "invent")
    new_skel = s3._get_skeleton_file()
    assert new_skel is not None
    assert Path(new_skel).exists()
    assert Path(new_skel).stat().st_size > 0

    latest_md5 = (
        requests.get(s3.TF_SKEL_URL(release="master", ext="md5", allow_redirects=True))
        .content.decode()
        .split()[0]
    )
    monkeypatch.setattr(s3, "TF_SKEL_MD5", latest_md5)
    assert s3._get_skeleton_file() is None

    monkeypatch.setattr(s3, "TF_SKEL_MD5", local_md5)
    monkeypatch.setattr(s3, "TF_SKEL_URL", "http://weird/{release}/{ext}".format)
    assert s3._get_skeleton_file() is None

    monkeypatch.setattr(
        s3, "TF_SKEL_URL", s3.TF_SKEL_URL(release="{release}", ext="{ext}z").format
    )
    assert s3._get_skeleton_file() is None


def test_update_s3(tmp_path, monkeypatch):
    """Exercise updating the S3 skeleton."""
    newhome = tmp_path / "templateflow"
    assert s3.update(newhome)
    assert not s3.update(newhome, overwrite=False)
    for p in (newhome / "tpl-MNI152NLin6Sym").glob("*.nii.gz"):
        p.unlink()
    assert s3.update(newhome, overwrite=False)

    # This should cover the remote zip file fetching
    monkeypatch.setattr(s3, "TF_SKEL_MD5", "invent")
    assert s3.update(newhome, local=False)
    assert not s3.update(newhome, local=False, overwrite=False)
    for p in (newhome / "tpl-MNI152NLin6Sym").glob("*.nii.gz"):
        p.unlink()
    assert s3.update(newhome, local=False, overwrite=False)



"""Configuration and settings."""
from os import getenv
from warnings import warn
from pathlib import Path

TF_DEFAULT_HOME = Path.home() / ".cache" / "templateflow"
TF_HOME = Path(getenv("TEMPLATEFLOW_HOME", str(TF_DEFAULT_HOME)))
TF_GITHUB_SOURCE = "https://github.com/templateflow/templateflow.git"
TF_S3_ROOT = "https://templateflow.s3.amazonaws.com"
TF_USE_DATALAD = getenv("TEMPLATEFLOW_USE_DATALAD", "false").lower() in (
    "true",
    "on",
    "1",
    "yes",
    "y",
)
TF_CACHED = True

if not TF_HOME.exists() or not list(TF_HOME.iterdir()):
    TF_CACHED = False
    warn(
        """\
TemplateFlow: repository not found at %s. Populating a new TemplateFlow stub.
If the path reported above is not the desired location for TemplateFlow, \
please set the TEMPLATEFLOW_HOME environment variable.\
"""
        % TF_HOME,
        ResourceWarning,
    )
    if TF_USE_DATALAD:
        try:
            from datalad.api import install
        except ImportError:
            TF_USE_DATALAD = False
        else:
            TF_HOME.parent.mkdir(exist_ok=True, parents=True)
            install(path=str(TF_HOME), source=TF_GITHUB_SOURCE, recursive=True)

    if not TF_USE_DATALAD:
        from ._s3 import update as _update_s3

        _update_s3(TF_HOME, local=True, overwrite=True)


def update(local=False, overwrite=True, silent=False):
    """Update an existing DataLad or S3 home."""
    if TF_USE_DATALAD and _update_datalad():
        return True

    from ._s3 import update as _update_s3

    return _update_s3(TF_HOME, local=local, overwrite=overwrite, silent=silent)


def setup_home(force=False):
    """Initialize/update TF's home if necessary."""
    if not force and not TF_CACHED:
        print(
            f"""\
TemplateFlow was not cached (TEMPLATEFLOW_HOME={TF_HOME}), \
a fresh initialization was done."""
        )
        return False
    return update(local=True, overwrite=False)


def _update_datalad():
    from datalad.api import update

    print("Updating TEMPLATEFLOW_HOME using DataLad ...")
    try:
        update(dataset=str(TF_HOME), recursive=True, merge=True)
    except Exception as e:
        warn(f"Error updating TemplateFlow's home directory (using DataLad): {e}")
    return True


TF_LAYOUT = None
try:
    from .bids import Layout

    TF_LAYOUT = Layout(
        TF_HOME,
        validate=False,
        config="templateflow",
        ignore=[
            ".git",
            ".datalad",
            ".gitannex",
            ".gitattributes",
            ".github",
            "scripts",
        ],
    )
except ImportError:
    pass



"""Extending pyBIDS for querying TemplateFlow."""
from pkg_resources import resource_filename
from bids.layout import BIDSLayout, add_config_paths

add_config_paths(templateflow=resource_filename("templateflow", "conf/config.json"))


class Layout(BIDSLayout):
    def __repr__(self):
        # A tidy summary of key properties
        s = """\
TemplateFlow Layout
 - Home: {}
 - Templates: {}.""".format(
            self.root, ", ".join(sorted(self.get_templates()))
        )
        return s



import pytest

from .. import api


# test setup to avoid cluttering pytest parameterize
mni2009_urls = [
    "https://doi.org/10.1016/j.neuroimage.2010.07.033",
    "https://doi.org/10.1016/S1053-8119(09)70884-5",
    "http://nist.mni.mcgill.ca/?p=904",
    "https://doi.org/10.1007/3-540-48714-X_16",
]

mni2009_fbib = (
    "@article{mni152nlin2009casym1,\n"
    "\tdoi = {10.1016/j.neuroimage.2010.07.033},\n"
    "\turl = {https://doi.org/10.1016%2Fj.neuroimage.2010.07.033},\n"
    "\tyear = 2011,\n"
    "\tmonth = {jan},\n"
    "\tpublisher = {Elsevier {BV}},\n"
    "\tvolume = {54},\n"
    "\tnumber = {1},\n"
    "\tpages = {313--327},\n"
    "\tauthor = {Vladimir Fonov and Alan C. Evans and Kelly Botteron and C. "
    "Robert Almli and Robert C. McKinstry and D. Louis Collins},\n"
    "\ttitle = {Unbiased average age-appropriate atlases for pediatric studies},\n"
    "\tjournal = {{NeuroImage}}\n}"
)

mni2009_lbib = (
    "@incollection{mni152nlin2009casym4,\n"
    "\tdoi = {10.1007/3-540-48714-x_16},\n"
    "\turl = {https://doi.org/10.1007%2F3-540-48714-x_16},\n"
    "\tyear = 1999,\n"
    "\tpublisher = {Springer Berlin Heidelberg},\n"
    "\tpages = {210--223},\n"
    "\tauthor = {D. Louis Collins and Alex P. Zijdenbos and Wim F. C. "
    "Baar{\\'{e}} and Alan C. Evans},\n"
    "\ttitle = {{ANIMAL}$\\mathplus${INSECT}: Improved Cortical Structure "
    "Segmentation},\n"
    "\tbooktitle = {Lecture Notes in Computer Science}\n}"
)

fslr_urls = [
    "https://doi.org/10.1093/cercor/bhr291",
    "https://github.com/Washington-University/HCPpipelines/tree/master/global/templates",
]

fslr_fbib = (
    "@article{fslr1,\n"
    "\tdoi = {10.1093/cercor/bhr291},\n"
    "\turl = {https://doi.org/10.1093%2Fcercor%2Fbhr291},\n"
    "\tyear = 2011,\n"
    "\tmonth = {nov},\n"
    "\tpublisher = {Oxford University Press ({OUP})},\n"
    "\tvolume = {22},\n"
    "\tnumber = {10},\n"
    "\tpages = {2241--2262},\n"
    "\tauthor = {D. C. Van Essen and M. F. Glasser and D. L. Dierker and J. "
    "Harwell and T. Coalson},\n"
    "\ttitle = {Parcellations and Hemispheric Asymmetries of Human Cerebral "
    "Cortex Analyzed on Surface-Based Atlases},\n"
    "\tjournal = {Cerebral Cortex}\n}"
)

fslr_lbib = (
    "https://github.com/Washington-University/HCPpipelines/tree/master/global/templates"
)


@pytest.mark.parametrize(
    "template,urls,fbib,lbib",
    [
        ("MNI152NLin2009cAsym", mni2009_urls, mni2009_fbib, mni2009_lbib),
        ("fsLR", fslr_urls, fslr_fbib, fslr_lbib),
        ("fsaverage", [], None, None),
    ],
)
def test_citations(tmp_path, template, urls, fbib, lbib):
    assert api.get_citations(template) == urls
    bibs = api.get_citations(template, bibtex=True)
    if bibs:
        assert bibs[0] == fbib
        assert bibs[-1] == lbib
    else:
        # no citations currently
        assert template == "fsaverage"



"""TemplateFlow's Python Client."""
from json import loads
from pathlib import Path
import re
import sys

from .conf import TF_LAYOUT, TF_S3_ROOT, TF_USE_DATALAD


def get(template, **kwargs):
    """
    Fetch one file from one particular template.

    Parameters
    ----------
    template : str
        A template identifier (e.g., ``MNI152NLin2009cAsym``).

    Keyword Arguments
    -----------------
    resolution: int or None
        Index to an specific spatial resolution of the template.
    suffix : str or None
        BIDS suffix
    atlas : str or None
        Name of a particular atlas
    hemi : str or None
        Hemisphere
    space : str or None
        Space template is mapped to
    density : str or None
        Surface density
    desc : str or None
        Description field

    Examples
    --------
    >>> str(get('MNI152Lin', resolution=1, suffix='T1w'))  # doctest: +ELLIPSIS
    '.../tpl-MNI152Lin/tpl-MNI152Lin_res-01_T1w.nii.gz'

    >>> str(get('MNI152Lin', resolution=2, suffix='T1w'))  # doctest: +ELLIPSIS
    '.../tpl-MNI152Lin/tpl-MNI152Lin_res-02_T1w.nii.gz'

    >>> [str(p) for p in get(
    ...     'MNI152Lin', suffix='T1w')]  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    ['.../tpl-MNI152Lin/tpl-MNI152Lin_res-01_T1w.nii.gz',
     '.../tpl-MNI152Lin/tpl-MNI152Lin_res-02_T1w.nii.gz']

    >>> str(get('fsLR', space=None, hemi='L',
    ...         density='32k', suffix='sphere'))  # doctest: +ELLIPSIS
    '.../tpl-fsLR_hemi-L_den-32k_sphere.surf.gii'

    """
    out_file = [
        Path(p) for p in TF_LAYOUT.get(template=template, return_type="file", **kwargs)
    ]

    # Try DataLad first
    dl_missing = [p for p in out_file if not p.is_file()]
    if TF_USE_DATALAD and dl_missing:
        for filepath in dl_missing:
            _datalad_get(filepath)
            dl_missing.remove(filepath)

    # Fall-back to S3 if some files are still missing
    s3_missing = [p for p in out_file if p.is_file() and p.stat().st_size == 0]
    for filepath in s3_missing + dl_missing:
        _s3_get(filepath)

    not_fetched = [str(p) for p in out_file if not p.is_file() or p.stat().st_size == 0]

    if not_fetched:
        msg = "Could not fetch template files: %s." % ", ".join(not_fetched)
        if dl_missing and not TF_USE_DATALAD:
            msg += (
                """\
The $TEMPLATEFLOW_HOME folder %s seems to contain an initiated DataLad \
dataset, but the environment variable $TEMPLATEFLOW_USE_DATALAD is not \
set or set to one of (false, off, 0). Please set $TEMPLATEFLOW_USE_DATALAD \
on (possible values: true, on, 1)."""
                % TF_LAYOUT.root
            )

        if s3_missing and TF_USE_DATALAD:
            msg += (
                """\
The $TEMPLATEFLOW_HOME folder %s seems to contain an plain \
dataset, but the environment variable $TEMPLATEFLOW_USE_DATALAD is \
set to one of (true, on, 1). Please set $TEMPLATEFLOW_USE_DATALAD \
off (possible values: false, off, 0)."""
                % TF_LAYOUT.root
            )

        raise RuntimeError(msg)

    if len(out_file) == 1:
        return out_file[0]
    return out_file


def templates(**kwargs):
    """
    Returns a list of available templates.

    Keyword Arguments
    -----------------
    resolution: int or None
        Index to an specific spatial resolution of the template.
    suffix : str or None
        BIDS suffix
    atlas : str
        Name of a particular atlas
    desc : str
        Description field

    Examples
    --------
    >>> base = ['MNI152Lin', 'MNI152NLin2009cAsym', 'NKI', 'OASIS30ANTs']
    >>> tpls = templates()
    >>> all([t in tpls for t in base])
    True

    >>> templates(suffix='PD')
    ['MNI152Lin', 'MNI152NLin2009cAsym', 'MNI152NLin2009cSym', 'MNIInfant', 'MNIPediatricAsym']

    """
    return sorted(TF_LAYOUT.get_templates(**kwargs))


def get_metadata(template):
    """
    Fetch one file from one template.

    Parameters
    ----------
    template : str
        A template identifier (e.g., ``MNI152NLin2009cAsym``).

    Examples
    --------
    >>> get_metadata('MNI152Lin')['Name']
    'Linear ICBM Average Brain (ICBM152) Stereotaxic Registration Model'

    """

    tf_home = Path(TF_LAYOUT.root)
    filepath = tf_home / ("tpl-%s" % template) / "template_description.json"

    # Ensure that template is installed and file is available
    if not filepath.is_file():
        _datalad_get(filepath)
    return loads(filepath.read_text())


def get_citations(template, bibtex=False):
    """
    Fetch template citations

    Parameters
    ----------
    template : :obj:`str`
        A template identifier (e.g., ``MNI152NLin2009cAsym``).
    bibtex : :obj:`bool`, optional
        Generate citations in BibTeX format.

    """
    data = get_metadata(template)
    refs = data.get("ReferencesAndLinks", [])
    if isinstance(refs, dict):
        refs = [x for x in refs.values()]

    if not bibtex:
        return refs

    return [_to_bibtex(ref, template, idx).rstrip() for idx, ref in enumerate(refs, 1)]


def _datalad_get(filepath):
    if not filepath:
        return

    from datalad import api
    from datalad.support.exceptions import IncompleteResultsError

    try:
        api.get(str(filepath))
    except IncompleteResultsError as exc:
        if exc.failed[0]["message"] == "path not associated with any dataset":
            from .conf import TF_GITHUB_SOURCE

            api.install(path=TF_LAYOUT.root, source=TF_GITHUB_SOURCE, recursive=True)
            api.get(str(filepath))
        else:
            raise


def _s3_get(filepath):
    from sys import stderr
    from math import ceil
    from tqdm import tqdm
    import requests

    path = str(filepath.relative_to(TF_LAYOUT.root))
    url = "%s/%s" % (TF_S3_ROOT, path)

    print("Downloading %s" % url, file=stderr)
    # Streaming, so we can iterate over the response.
    r = requests.get(url, stream=True)

    # Total size in bytes.
    total_size = int(r.headers.get("content-length", 0))
    block_size = 1024
    wrote = 0
    if not filepath.is_file():
        filepath.unlink()

    with filepath.open("wb") as f:
        for data in tqdm(
            r.iter_content(block_size),
            total=ceil(total_size // block_size),
            unit="B",
            unit_scale=True,
        ):
            wrote = wrote + len(data)
            f.write(data)
    if total_size != 0 and wrote != total_size:
        raise RuntimeError("ERROR, something went wrong")


def _to_bibtex(doi, template, idx):
    try:
        from doi2bib.crossref import get_bib_from_doi
    except ImportError:
        print(
            "Cannot generate BibTeX citation, missing doi2bib dependency",
            file=sys.stderr,
        )
        return doi

    if "doi.org" not in doi:
        return doi
    bib = get_bib_from_doi(doi)[1]
    # replace identifier with template name
    m = re.search(r"([A-Z])\w+", bib)
    return bib.replace(m.group(), "%s%s" % (template.lower(), idx))



#!/usr/bin/env python
"""Templateflow's setup script."""
import sys
from setuptools import setup


# Give setuptools a hint to complain if it's too old a version
# 30.3.0 allows us to put most metadata in setup.cfg
# Should match pyproject.toml
SETUP_REQUIRES = [
    "setuptools >= 42.0",
    "setuptools_scm >= 3.4",
    "setuptools_scm_git_archive",
    "toml",
]
# This enables setuptools to install wheel on-the-fly
SETUP_REQUIRES += ["wheel"] if "bdist_wheel" in sys.argv else []

if __name__ == "__main__":
    """ Install entry-point """
    setup(
        name="templateflow",
        setup_requires=SETUP_REQUIRES,
    )



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import pytest
from shutil import rmtree
from importlib import reload
from templateflow import conf as tfc


@pytest.mark.parametrize("use_datalad", ["off", "on"])
def test_conf_init(monkeypatch, tmp_path, capsys, use_datalad):
    """Check the correct functioning of config set-up."""
    home = (tmp_path / "-".join(("tf", "dl", use_datalad))).resolve()
    monkeypatch.setenv("TEMPLATEFLOW_USE_DATALAD", use_datalad)
    monkeypatch.setenv("TEMPLATEFLOW_HOME", str(home))

    # First execution, the S3 stub is created (or datalad install)
    reload(tfc)
    assert tfc.TF_CACHED is False
    assert str(tfc.TF_HOME) == str(home)

    reload(tfc)
    assert tfc.TF_CACHED is True
    assert str(tfc.TF_HOME) == str(home)


@pytest.mark.parametrize("use_datalad", ["off", "on"])
def test_setup_home(monkeypatch, tmp_path, capsys, use_datalad):
    """Check the correct functioning of the installation hook."""
    home = (tmp_path / "-".join(("tf", "dl", use_datalad))).resolve()
    monkeypatch.setenv("TEMPLATEFLOW_USE_DATALAD", use_datalad)
    monkeypatch.setenv("TEMPLATEFLOW_HOME", str(home))

    reload(tfc)
    # First execution, the S3 stub is created (or datalad install)
    assert tfc.TF_CACHED is False
    assert tfc.setup_home() is False
    out = capsys.readouterr()[0]
    assert out.startswith("TemplateFlow was not cached")
    assert ("TEMPLATEFLOW_HOME=%s" % home) in out
    assert home.exists()
    assert len(list(home.iterdir())) > 0

    updated = tfc.setup_home(force=True)  # Templateflow is now cached
    out = capsys.readouterr()[0]
    assert not out.startswith("TemplateFlow was not cached")

    if use_datalad == "on":
        assert out.startswith("Updating TEMPLATEFLOW_HOME using DataLad")
        assert updated is True

    elif use_datalad == "off":
        # At this point, S3 should be up-to-date
        assert updated is False
        assert out.startswith("TEMPLATEFLOW_HOME directory (S3 type) was up-to-date.")

        # Let's force an update
        rmtree(str(home / "tpl-MNI152NLin2009cAsym"))
        updated = tfc.setup_home(force=True)
        out = capsys.readouterr()[0]
        assert updated is True
        assert out.startswith("Updating TEMPLATEFLOW_HOME using S3.")

    reload(tfc)
    assert tfc.TF_CACHED is True
    updated = tfc.setup_home()  # Templateflow is now cached
    out = capsys.readouterr()[0]
    assert not out.startswith("TemplateFlow was not cached")

    if use_datalad == "on":
        assert out.startswith("Updating TEMPLATEFLOW_HOME using DataLad")
        assert updated is True

    elif use_datalad == "off":
        # At this point, S3 should be up-to-date
        assert updated is False
        assert out.startswith("TEMPLATEFLOW_HOME directory (S3 type) was up-to-date.")

        # Let's force an update
        rmtree(str(home / "tpl-MNI152NLin2009cAsym"))
        updated = tfc.setup_home()
        out = capsys.readouterr()[0]
        assert updated is True
        assert out.startswith("Updating TEMPLATEFLOW_HOME using S3.")


def test_layout(monkeypatch, tmp_path):
    monkeypatch.setenv("TEMPLATEFLOW_USE_DATALAD", "off")

    lines = ("%s" % tfc.TF_LAYOUT).splitlines()
    assert lines[0] == "TemplateFlow Layout"
    assert lines[1] == " - Home: %s" % tfc.TF_HOME
    assert lines[2].startswith(" - Templates:")



"""Tooling to handle S3 downloads."""
from pathlib import Path
from tempfile import mkstemp
from pkg_resources import resource_filename

TF_SKEL_URL = (
    "https://raw.githubusercontent.com/templateflow/python-client/"
    "{release}/templateflow/conf/templateflow-skel.{ext}"
).format
TF_SKEL_PATH = Path(resource_filename("templateflow", "conf/templateflow-skel.zip"))
TF_SKEL_MD5 = Path(
    resource_filename("templateflow", "conf/templateflow-skel.md5")
).read_text()


def update(dest, local=True, overwrite=True, silent=False):
    """Update an S3-backed TEMPLATEFLOW_HOME repository."""
    skel_file = Path((_get_skeleton_file() if not local else None) or TF_SKEL_PATH)

    retval = _update_skeleton(skel_file, dest, overwrite=overwrite, silent=silent)
    if skel_file != TF_SKEL_PATH:
        skel_file.unlink()
    return retval


def _get_skeleton_file():
    import requests

    try:
        r = requests.get(TF_SKEL_URL(release="master", ext="md5", allow_redirects=True))
    except requests.exceptions.ConnectionError:
        return

    if not r.ok:
        return

    if r.content.decode().split()[0] != TF_SKEL_MD5:
        r = requests.get(TF_SKEL_URL(release="master", ext="zip", allow_redirects=True))
        if r.ok:
            from os import close

            fh, skel_file = mkstemp(suffix=".zip")
            Path(skel_file).write_bytes(r.content)
            close(fh)
            return skel_file


def _update_skeleton(skel_file, dest, overwrite=True, silent=False):
    from zipfile import ZipFile

    dest = Path(dest)
    dest.mkdir(exist_ok=True, parents=True)
    with ZipFile(skel_file, "r") as zipref:
        if overwrite:
            zipref.extractall(str(dest))
            return True

        allfiles = zipref.namelist()
        current_files = [s.relative_to(dest) for s in dest.glob("**/*")]
        existing = sorted(set(["%s/" % s.parent for s in current_files])) + [
            str(s) for s in current_files
        ]
        newfiles = sorted(set(allfiles) - set(existing))
        if newfiles:
            if not silent:
                print(
                    "Updating TEMPLATEFLOW_HOME using S3. Adding:\n%s" % '\n'.join(newfiles)
                )
            zipref.extractall(str(dest), members=newfiles)
            return True
    if not silent:
        print("TEMPLATEFLOW_HOME directory (S3 type) was up-to-date.")
    return False
